# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - EventAggregator library

    @copyright: 2008, 2009, 2010, 2011 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
                2005-2008 MoinMoin:ThomasWaldmann.
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin import search, version
from MoinMoin import wikiutil
import calendar
import datetime
import time
import re
import bisect
import operator

try:
    set
except NameError:
    from sets import Set as set

try:
    import pytz
except ImportError:
    pytz = None

escape = wikiutil.escape

__version__ = "0.7"

# Date labels.

month_labels = ["January", "February", "March", "April", "May", "June",
    "July", "August", "September", "October", "November", "December"]
weekday_labels = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]

# Regular expressions where MoinMoin does not provide the required support.

category_regexp = None

# Page parsing.

definition_list_regexp = re.compile(ur'(?P<wholeterm>^(?P<optcomment>#*)\s+(?P<term>.*?):: )(?P<desc>.*?)$', re.UNICODE | re.MULTILINE)
category_membership_regexp = re.compile(ur"^\s*((Category\S+)(\s+Category\S+)*)\s*$", re.MULTILINE | re.UNICODE)

# Value parsing.

country_code_regexp = re.compile(ur'(?:^|\W)(?P<code>[A-Z]{2})(?:$|\W+$)', re.UNICODE)

month_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})'
date_regexp_str = ur'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
time_regexp_str = ur'(?P<hour>[0-2][0-9]):(?P<minute>[0-5][0-9])(?::(?P<second>[0-6][0-9]))?'
timezone_offset_str = ur'(?P<offset>(UTC)?(?:(?P<sign>[-+])(?P<hours>[0-9]{2})(?::?(?P<minutes>[0-9]{2}))?))'
timezone_olson_str = ur'(?P<olson>[a-zA-Z]+(?:/[-_a-zA-Z]+){1,2})'
timezone_utc_str = ur'UTC'
timezone_regexp_str = ur'(?P<zone>' + timezone_offset_str + '|' + timezone_olson_str + '|' + timezone_utc_str + ')'
datetime_regexp_str = date_regexp_str + ur'(?:\s+' + time_regexp_str + ur'(?:\s+' + timezone_regexp_str + ur')?)?'

month_regexp = re.compile(month_regexp_str, re.UNICODE)
date_regexp = re.compile(date_regexp_str, re.UNICODE)
time_regexp = re.compile(time_regexp_str, re.UNICODE)
datetime_regexp = re.compile(datetime_regexp_str, re.UNICODE)
timezone_olson_regexp = re.compile(timezone_olson_str, re.UNICODE)
timezone_offset_regexp = re.compile(timezone_offset_str, re.UNICODE)

verbatim_regexp = re.compile(ur'(?:'
    ur'<<Verbatim\((?P<verbatim>.*?)\)>>'
    ur'|'
    ur'\[\[Verbatim\((?P<verbatim2>.*?)\)\]\]'
    ur'|'
    ur'`(?P<monospace>.*?)`'
    ur'|'
    ur'{{{(?P<preformatted>.*?)}}}'
    ur')', re.UNICODE)

# Utility functions.

def getCategoryPattern(request):
    global category_regexp

    try:
        return request.cfg.cache.page_category_regexact
    except AttributeError:

        # Use regular expression from MoinMoin 1.7.1 otherwise.

        if category_regexp is None:
            category_regexp = re.compile(u'^%s$' % ur'(?P<all>Category(?P<key>(?!Template)\S+))', re.UNICODE)
        return category_regexp

def int_or_none(x):
    if x is None:
        return x
    else:
        return int(x)

def sort_none_first(x, y):
    if x is None:
        return -1
    elif y is None:
        return 1
    else:
        return cmp(x, y)

# Utility classes and associated functions.

class Form:

    """
    A wrapper preserving MoinMoin 1.8.x (and earlier) behaviour in a 1.9.x
    environment.
    """

    def __init__(self, form):
        self.form = form

    def get(self, name, default=None):
        values = self.form.getlist(name)
        if not values:
            return default
        else:
            return values

    def __getitem__(self, name):
        return self.form.getlist(name)

class ActionSupport:

    """
    Work around disruptive MoinMoin changes in 1.9, and also provide useful
    convenience methods.
    """

    def get_form(self):
        return get_form(self.request)

    def _get_selected(self, value, input_value):

        """
        Return the HTML attribute text indicating selection of an option (or
        otherwise) if 'value' matches 'input_value'.
        """

        return input_value is not None and value == input_value and 'selected="selected"' or ''

    def _get_selected_for_list(self, value, input_values):

        """
        Return the HTML attribute text indicating selection of an option (or
        otherwise) if 'value' matches one of the 'input_values'.
        """

        return value in input_values and 'selected="selected"' or ''

    def _get_input(self, form, name, default=None):

        """
        Return the input from 'form' having the given 'name', returning either
        the input converted to an integer or the given 'default' (optional, None
        if not specified).
        """

        value = form.get(name, [None])[0]
        if not value: # true if 0 obtained
            return default
        else:
            return int(value)

    def get_month_lists(self, default_as_current=0):

        """
        Return two lists of HTML element definitions corresponding to the start
        and end month selection controls, with months selected according to any
        values that have been specified via request parameters.
        """

        _ = self._
        form = self.get_form()

        # Initialise month lists.

        start_month_list = []
        end_month_list = []

        start_month = self._get_input(form, "start-month", default_as_current and getCurrentMonth().month() or None)
        end_month = self._get_input(form, "end-month", start_month)

        # Prepare month lists, selecting specified months.

        if not default_as_current:
            start_month_list.append('<option value=""></option>')
            end_month_list.append('<option value=""></option>')

        for month in range(1, 13):
            month_label = escape(_(getMonthLabel(month)))
            selected = self._get_selected(month, start_month)
            start_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))
            selected = self._get_selected(month, end_month)
            end_month_list.append('<option value="%02d" %s>%s</option>' % (month, selected, month_label))

        return start_month_list, end_month_list

    def get_year_defaults(self, default_as_current=0):

        "Return defaults for the start and end years."

        form = self.get_form()

        start_year_default = form.get("start-year", [default_as_current and getCurrentYear() or ""])[0]
        end_year_default = form.get("end-year", [default_as_current and start_year_default or ""])[0]

        return start_year_default, end_year_default

    def get_day_defaults(self, default_as_current=0):

        "Return defaults for the start and end days."

        form = self.get_form()

        start_day_default = form.get("start-day", [default_as_current and getCurrentDate().day() or ""])[0]
        end_day_default = form.get("end-day", [default_as_current and start_day_default or ""])[0]

        return start_day_default, end_day_default

def get_form(request):

    "Work around disruptive MoinMoin changes in 1.9."

    if hasattr(request, "values"):
        return Form(request.values)
    else:
        return request.form

class send_headers:

    """
    A wrapper to preserve MoinMoin 1.8.x (and earlier) request behaviour in a
    1.9.x environment.
    """

    def __init__(self, request):
        self.request = request

    def __call__(self, headers):
        for header in headers:
            parts = header.split(":")
            self.request.headers.add(parts[0], ":".join(parts[1:]))

def escattr(s):
    return escape(s, 1)

# Textual representations.

def getHTTPTimeString(tmtuple):
    return "%s, %02d %s %04d %02d:%02d:%02d GMT" % (
        getDayLabel(tmtuple.tm_wday),
        tmtuple.tm_mday,
        getMonthLabel(tmtuple.tm_mon),
        tmtuple.tm_year,
        tmtuple.tm_hour,
        tmtuple.tm_min,
        tmtuple.tm_sec
        )

def getSimpleWikiText(text):

    """
    Return the plain text representation of the given 'text' which may employ
    certain Wiki syntax features, such as those providing verbatim or monospaced
    text.
    """

    # NOTE: Re-implementing support for verbatim text and linking avoidance.

    return "".join([s for s in verbatim_regexp.split(text) if s is not None])

def getEncodedWikiText(text):

    "Encode the given 'text' in a verbatim representation."

    return "<<Verbatim(%s)>>" % text

def getPrettyTitle(title):

    "Return a nicely formatted version of the given 'title'."

    return title.replace("_", " ").replace("/", u"  ")

def getMonthLabel(month):

    "Return an unlocalised label for the given 'month'."

    return month_labels[month - 1] # zero-based labels

def getDayLabel(weekday):

    "Return an unlocalised label for the given 'weekday'."

    return weekday_labels[weekday]

# Action support functions.

def getPageRevision(page):

    "Return the revision details dictionary for the given 'page'."

    # From Page.edit_info...

    if hasattr(page, "editlog_entry"):
        line = page.editlog_entry()
    else:
        line = page._last_edited(page.request) # MoinMoin 1.5.x and 1.6.x

    # Similar to Page.mtime_usecs behaviour...

    if line:
        timestamp = line.ed_time_usecs
        mtime = wikiutil.version2timestamp(long(timestamp)) # must be long for py 2.2.x
        comment = line.comment
    else:
        mtime = 0
        comment = ""

    return {"timestamp" : time.gmtime(mtime), "comment" : comment}

# Category discovery and searching.

def getCategories(request):

    """
    From the AdvancedSearch macro, return a list of category page names using
    the given 'request'.
    """

    # This will return all pages with "Category" in the title.

    cat_filter = getCategoryPattern(request).search
    return request.rootpage.getPageList(filter=cat_filter)

def getCategoryMapping(category_pagenames, request):

    """
    For the given 'category_pagenames' return a list of tuples of the form
    (category name, category page name) using the given 'request'.
    """

    cat_pattern = getCategoryPattern(request)
    mapping = []
    for pagename in category_pagenames:
        name = cat_pattern.match(pagename).group("key")
        if name != "Category":
            mapping.append((name, pagename))
    mapping.sort()
    return mapping

def getCategoryPages(pagename, request):

    """
    Return the pages associated with the given category 'pagename' using the
    'request'.
    """

    query = search.QueryParser().parse_query('category:%s' % pagename)
    results = search.searchPages(request, query, "page_name")

    cat_pattern = getCategoryPattern(request)
    pages = []
    for page in results.hits:
        if not cat_pattern.match(page.page_name):
            pages.append(page)
    return pages

# The main activity functions.

class EventPage:

    "An event page."

    def __init__(self, page):
        self.page = page
        self.events = None
        self.body = None
        self.categories = None

    def copyPage(self, page):

        "Copy the body of the given 'page'."

        self.body = page.getBody()

    def getPageURL(self, request):

        "Using 'request', return the URL of this page."

        return request.getQualifiedURL(self.page.url(request, relative=0))

    def getFormat(self):

        "Get the format used on this page."

        return self.page.pi["format"]

    def getRevisions(self):

        "Return a list of page revisions."

        return self.page.getRevList()

    def getPageRevision(self):

        "Return the revision details dictionary for this page."

        return getPageRevision(self.page)

    def getPageName(self):

        "Return the page name."

        return self.page.page_name

    def getPrettyPageName(self):

        "Return a nicely formatted title/name for this page."

        return getPrettyPageName(self.page)

    def getBody(self):

        "Get the current page body."

        if self.body is None:
            self.body = self.page.get_raw_body()
        return self.body

    def getEvents(self):

        "Return a list of events from this page."

        if self.events is None:
            details = {}
            self.events = [Event(self, details)]

            if self.getFormat() == "wiki":
                for match in definition_list_regexp.finditer(self.getBody()):

                    # Skip commented-out items.

                    if match.group("optcomment"):
                        continue

                    # Permit case-insensitive list terms.

                    term = match.group("term").lower()
                    desc = match.group("desc")

                    # Special value type handling.

                    # Dates.

                    if term in ("start", "end"):
                        desc = getDateTime(desc)

                    # Lists (whose elements may be quoted).

                    elif term in ("topics", "categories"):
                        desc = [getSimpleWikiText(value.strip()) for value in desc.split(",") if value.strip()]

                    # Labels which may well be quoted.

                    elif term in ("title", "summary", "description", "location"):
                        desc = getSimpleWikiText(desc)

                    if desc is not None:

                        # Handle apparent duplicates by creating a new set of
                        # details.

                        if details.has_key(term):

                            # Make a new event.

                            details = {}
                            self.events.append(Event(self, details))

                        details[term] = desc

        return self.events

    def setEvents(self, events):

        "Set the given 'events' on this page."

        self.events = events

    def getCategoryMembership(self):

        "Get the category names from this page."

        if self.categories is None:
            body = self.getBody()
            match = category_membership_regexp.search(body)
            self.categories = match.findall().split()

        return self.categories

    def setCategoryMembership(self, category_names):

        """
        Set the category membership for the page using the specified
        'category_names'.
        """

        self.categories = category_names

    def flushEventDetails(self):

        "Flush the current event details to this page's body text."

        new_body_parts = []
        end_of_last_match = 0
        body = self.getBody()

        events = iter(self.getEvents())

        event = events.next()
        event_details = event.getDetails()
        replaced_terms = set()

        for match in definition_list_regexp.finditer(body):

            # Permit case-insensitive list terms.

            term = match.group("term").lower()
            desc = match.group("desc")

            # Check that the term has not already been substituted. If so,
            # get the next event.

            if term in replaced_terms:
                try:
                    event = events.next()

                # No more events.

                except StopIteration:
                    break

                event_details = event.getDetails()
                replaced_terms = set()

            # Add preceding text to the new body.

            new_body_parts.append(body[end_of_last_match:match.start()])

            # Get the matching regions, adding the term to the new body.

            new_body_parts.append(match.group("wholeterm"))

            # Special value type handling.

            if event_details.has_key(term):

                # Dates.

                if term in ("start", "end"):
                    desc = desc.replace("YYYY-MM-DD", str(event_details[term]))

                # Lists (whose elements may be quoted).

                elif term in ("topics", "categories"):
                    desc = ", ".join([getEncodedWikiText(item) for item in event_details[term]])

                # Labels which must be quoted.

                elif term in ("title", "summary"):
                    desc = getEncodedWikiText(event_details[term])

                # Text which need not be quoted, but it will be Wiki text.

                elif term in ("description", "link", "location"):
                    desc = event_details[term]

                replaced_terms.add(term)

            # Add the replaced value.

            new_body_parts.append(desc)

            # Remember where in the page has been processed.

            end_of_last_match = match.end()

        # Write the rest of the page.

        new_body_parts.append(body[end_of_last_match:])

        self.body = "".join(new_body_parts)

    def flushCategoryMembership(self):

        "Flush the category membership to the page body."

        body = self.getBody()
        category_names = self.getCategoryMembership()
        match = category_membership_regexp.search(body)

        if match:
            self.body = "".join([body[:match.start()], " ".join(category_names), body[match.end():]])

    def saveChanges(self):

        "Save changes to the event."

        self.flushEventDetails()
        self.flushCategoryMembership()
        self.page.saveText(self.getBody(), 0)

    def linkToPage(self, request, text, query_string=None):

        """
        Using 'request', return a link to this page with the given link 'text'
        and optional 'query_string'.
        """

        return linkToPage(request, self.page, text, query_string)

class ActsAsTimespan:
    pass

class Event(ActsAsTimespan):

    "A description of an event."

    def __init__(self, page, details):
        self.page = page
        self.details = details

    def __hash__(self):
        return hash(self.getSummary())

    def getPage(self):

        "Return the page describing this event."

        return self.page

    def setPage(self, page):

        "Set the 'page' describing this event."

        self.page = page

    def getSummary(self, event_parent=None):

        """
        Return either the given title or summary of the event according to the
        event details, or a summary made from using the pretty version of the
        page name.

        If the optional 'event_parent' is specified, any page beneath the given
        'event_parent' page in the page hierarchy will omit this parent information
        if its name is used as the summary.
        """

        event_details = self.details

        if event_details.has_key("title"):
            return event_details["title"]
        elif event_details.has_key("summary"):
            return event_details["summary"]
        else:
            # If appropriate, remove the parent details and "/" character.

            title = self.page.getPageName()

            if event_parent and title.startswith(event_parent):
                title = title[len(event_parent.rstrip("/")) + 1:]

            return getPrettyTitle(title)

    def getDetails(self):

        "Return the details for this event."

        return self.details

    def setDetails(self, event_details):

        "Set the 'event_details' for this event."

        self.details = event_details

    # Timespan-related methods.

    def __contains__(self, other):
        return self == other

    def __cmp__(self, other):
        if isinstance(other, Event):
            return cmp(self.as_timespan(), other.as_timespan())
        else:
            return cmp(self.as_timespan(), other)

    def as_timespan(self):
        details = self.details
        if details.has_key("start") and details.has_key("end"):
            return Timespan(details["start"], details["end"])
        else:
            return None

    def as_limits(self):
        return self.as_timespan().as_limits()

def getEvents(request, category_names, calendar_start=None, calendar_end=None, resolution="month"):

    """
    Using the 'request', generate a list of events found on pages belonging to
    the specified 'category_names', using the optional 'calendar_start' and
    'calendar_end' values to indicate a window of interest.

    The optional 'resolution' determines the unit of time used in providing the
    results:

      * a list of events
      * a dictionary mapping time units to event lists (within the window of
        interest), usable as a kind of index to groups of events
      * a list of all events within the window of interest
      * the earliest time value of an event within the window of interest
      * the latest time value of an event within the window of interest.
    """

    # Dates need to comply with the requested resolution.
    # Here, None values need to be preserved when converting.

    if resolution == "month":
        convert = lambda x: x and x.as_month()
        get_values = lambda x, y: x.months_until(y)
    else:
        convert = lambda x: x and x.as_date()
        get_values = lambda x, y: x.days_until(y)

    # Re-order the window, if appropriate.

    if calendar_start is not None and calendar_end is not None and calendar_start > calendar_end:
        calendar_start, calendar_end = map(convert, (calendar_end, calendar_start))

    # Otherwise, just convert the calendar limits.

    else:
        calendar_start, calendar_end = map(convert, (calendar_start, calendar_end))

    calendar_period = Timespan(calendar_start, calendar_end)

    events = []
    shown_events = {}
    all_shown_events = []
    processed_pages = set()

    earliest = None
    latest = None

    for category_name in category_names:

        # Get the pages and page names in the category.

        pages_in_category = getCategoryPages(category_name, request)

        # Visit each page in the category.

        for page_in_category in pages_in_category:
            pagename = page_in_category.page_name

            # Only process each page once.

            if pagename in processed_pages:
                continue
            else:
                processed_pages.add(pagename)

            # Get a real page, not a result page.

            event_page = EventPage(Page(request, pagename))

            # Get all events described in the page.

            for event in event_page.getEvents():
                event_details = event.getDetails()

                # Remember the event.

                events.append(event)

                # Test for the suitability of the event.

                if event.as_timespan() is not None:
                    start, end = map(convert, event.as_timespan().as_limits())

                    # Compare the dates to the requested calendar window, if any.

                    if event in calendar_period:

                        all_shown_events.append(event)

                        if earliest is None or start < earliest:
                            earliest = start
                        if latest is None or end > latest:
                            latest = end

                        # Store the event in the time-specific dictionary.

                        first = max(start, calendar_start or start)
                        last = min(end, calendar_end or end)

                        for event_time_value in get_values(first, last):
                            if not shown_events.has_key(event_time_value):
                                shown_events[event_time_value] = []
                            shown_events[event_time_value].append(event)

    return events, shown_events, all_shown_events, earliest, latest

def setEventTimestamps(request, events):

    """
    Using 'request', set timestamp details in the details dictionary of each of
    the 'events'.

    Return the latest timestamp found.
    """

    latest = None

    for event in events:
        event_details = event.getDetails()
        event_page = event.getPage()

        # Get the initial revision of the page.

        revisions = event_page.getRevisions()
        event_page_initial = Page(request, event_page.getPageName(), rev=revisions[-1])

        # Get the created and last modified times.

        initial_revision = getPageRevision(event_page_initial)
        event_details["created"] = initial_revision["timestamp"]
        latest_revision = event_page.getPageRevision()
        event_details["last-modified"] = latest_revision["timestamp"]
        event_details["sequence"] = len(revisions) - 1
        event_details["last-comment"] = latest_revision["comment"]

        if latest is None or latest < event_details["last-modified"]:
            latest = event_details["last-modified"]

    return latest

def getOrderedEvents(events):

    """
    Return a list with the given 'events' ordered according to their start and
    end dates.
    """

    ordered_events = events[:]
    ordered_events.sort()
    return ordered_events

def getConcretePeriod(calendar_start, calendar_end, earliest, latest):

    """
    From the requested 'calendar_start' and 'calendar_end', which may be None,
    indicating that no restriction is imposed on the period for each of the
    boundaries, use the 'earliest' and 'latest' event months to define a
    specific period of interest.
    """

    # Define the period as starting with any specified start month or the
    # earliest event known, ending with any specified end month or the latest
    # event known.

    first = calendar_start or earliest
    last = calendar_end or latest

    # If there is no range of months to show, perhaps because there are no
    # events in the requested period, and there was no start or end month
    # specified, show only the month indicated by the start or end of the
    # requested period. If all events were to be shown but none were found show
    # the current month.

    if isinstance(first, Date):
        get_current = getCurrentDate
    else:
        get_current = getCurrentMonth

    if first is None:
        first = last or get_current()
    if last is None:
        last = first or get_current()

    # Permit "expiring" periods (where the start date approaches the end date).

    return min(first, last), last

def getCoverage(start, end, events, resolution="date"):

    """
    Within the period defined by the 'start' and 'end' dates, determine the
    coverage of the days in the period by the given 'events', returning a
    collection of timespans, along with a dictionary mapping locations to
    collections of slots, where each slot contains a tuple of the form
    (timespans, events).
    """

    all_events = {}
    full_coverage = TimespanCollection(resolution)
    coverage_period = full_coverage.convert(Timespan(start, end))

    # Get event details.

    for event in events:
        event_details = event.getDetails()

        # Test for the event in the period.

        if event in coverage_period:

            # Find the coverage of this period for the event.

            event_location = event_details.get("location")

            # Update the overall coverage.

            full_coverage.insert_in_order(event)

            # Add a new events list for a new location.
            # Locations can be unspecified, thus None refers to all unlocalised
            # events.

            if not all_events.has_key(event_location):
                all_events[event_location] = [TimespanCollection(resolution, [event])]

            # Try and fit the event into an events list.

            else:
                slot = all_events[event_location]

                for slot_events in slot:

                    # Where the event does not overlap with the events in the
                    # current collection, add it alongside these events.

                    if not event in slot_events:
                        slot_events.insert_in_order(event)
                        break

                # Make a new element in the list if the event cannot be
                # marked alongside existing events.

                else:
                    slot.append(TimespanCollection(resolution, [event]))

    return full_coverage, all_events

def getCoverageScale(coverage):

    """
    Return a scale for the given coverage so that the times involved are
    exposed. The scale consists of a list of non-overlapping timespans forming
    a contiguous period of time.
    """

    times = set()
    for timespan in coverage:
        start, end = timespan.as_limits()

        # Add either genuine times or dates converted to times.

        if isinstance(start, DateTime):
            times.add(start)
        else:
            times.add(start.as_datetime(None, None, None, None))

        if isinstance(end, DateTime):
            times.add(end)
        else:
            times.add(end.as_date().next_day())

    times = list(times)
    times.sort(cmp_dates_as_day_start)

    scale = []
    first = 1
    start = None
    for time in times:
        if not first:
            scale.append(Timespan(start, time))
        else:
            first = 0
        start = time

    return scale

# Date-related functions.

def cmp_dates_as_day_start(a, b):

    """
    Compare dates/datetimes 'a' and 'b' treating dates without time information
    as the earliest time in a particular day.
    """

    are_equal = a == b

    if are_equal:
        a2 = a.as_datetime_or_date()
        b2 = b.as_datetime_or_date()

        if isinstance(a2, Date) and isinstance(b2, DateTime):
            return -1
        elif isinstance(a2, DateTime) and isinstance(b2, Date):
            return 1

    return cmp(a, b)

class Period:

    "A simple period of time."

    def __init__(self, data):
        self.data = data

    def months(self):
        return self.data[0] * 12 + self.data[1]

class Temporal:

    "A simple temporal representation, common to dates and times."

    def __init__(self, data):
        self.data = list(data)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.data)

    def __hash__(self):
        return hash(self.as_tuple())

    def as_tuple(self):
        return tuple(self.data)

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is derived from a
        comparison of the instances' date(time) data at the highest common
        resolution, meaning that if a date is compared to a datetime, the
        datetime will be considered as a date. Thus, a date and a datetime
        referring to the same date will be considered equal.
        """

        if not isinstance(other, Temporal):
            return NotImplemented
        else:
            data = self.as_tuple()
            other_data = other.as_tuple()
            length = min(len(data), len(other_data))
            return cmp(data[:length], other_data[:length])

    def until(self, start, end, nextfn, prevfn):

        """
        Return a collection of units of time by starting from the given 'start'
        and stepping across intervening units until 'end' is reached, using the
        given 'nextfn' and 'prevfn' to step from one unit to the next.
        """

        current = start
        units = [current]
        if current < end:
            while current < end:
                current = nextfn(current)
                units.append(current)
        elif current > end:
            while current > end:
                current = prevfn(current)
                units.append(current)
        return units

    def ambiguous(self):

        "Only times can be ambiguous."

        return 0

class Month(Temporal):

    "A simple year-month representation."

    def __str__(self):
        return "%04d-%02d" % self.as_tuple()[:2]

    def as_datetime(self, day, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (day, hour, minute, second, zone))

    def as_date(self, day):
        return Date(self.as_tuple() + (day,))

    def as_month(self):
        return self

    def year(self):
        return self.data[0]

    def month(self):
        return self.data[1]

    def month_properties(self):

        """
        Return the weekday of the 1st of the month, along with the number of
        days, as a tuple.
        """

        year, month = self.as_tuple()[:2]
        return calendar.monthrange(year, month)

    def month_update(self, n=1):

        "Return the month updated by 'n' months."

        year, month = self.as_tuple()[:2]
        return Month((year + (month - 1 + n) / 12, (month - 1 + n) % 12 + 1))

    def next_month(self):

        "Return the month following this one."

        return self.month_update(1)

    def previous_month(self):

        "Return the month preceding this one."

        return self.month_update(-1)

    def __sub__(self, start):

        """
        Return the difference in years and months between this month and the
        'start' month as a period.
        """

        return Period([(x - y) for x, y in zip(self.data, start.data)])

    def months_until(self, end):

        "Return the collection of months from this month until 'end'."

        return self.until(self.as_month(), end.as_month(), Month.next_month, Month.previous_month)

class Date(Month):

    "A simple year-month-day representation."

    def constrain(self):
        year, month, day = self.as_tuple()[:3]

        month = max(min(month, 12), 1)
        wd, last_day = calendar.monthrange(year, month)
        day = max(min(day, last_day), 1)

        self.data[1:3] = month, day

    def __str__(self):
        return "%04d-%02d-%02d" % self.as_tuple()[:3]

    def as_datetime(self, hour, minute, second, zone):
        return DateTime(self.as_tuple() + (hour, minute, second, zone))

    def as_date(self):
        return self

    def as_datetime_or_date(self):
        return self

    def as_month(self):
        return Month(self.data[:2])

    def day(self):
        return self.data[2]

    def day_update(self, n=1):

        "Return the month updated by 'n' days."

        delta = datetime.timedelta(n)
        dt = datetime.date(*self.as_tuple()[:3])
        dt_new = dt + delta
        return Date((dt_new.year, dt_new.month, dt_new.day))

    def next_day(self):

        "Return the date following this one."

        year, month, day = self.as_tuple()[:3]
        _wd, end_day = calendar.monthrange(year, month)
        if day == end_day:
            if month == 12:
                return Date((year + 1, 1, 1))
            else:
                return Date((year, month + 1, 1))
        else:
            return Date((year, month, day + 1))

    def previous_day(self):

        "Return the date preceding this one."

        year, month, day = self.as_tuple()[:3]
        if day == 1:
            if month == 1:
                return Date((year - 1, 12, 31))
            else:
                _wd, end_day = calendar.monthrange(year, month - 1)
                return Date((year, month - 1, end_day))
        else:
            return Date((year, month, day - 1))

    def days_until(self, end):

        "Return the collection of days from this date until 'end'."

        return self.until(self.as_date(), end.as_date(), Date.next_day, Date.previous_day)

class DateTime(Date):

    "A simple date plus time representation."

    def constrain(self):
        Date.constrain(self)

        hour, minute, second = self.as_tuple()[3:6]

        if self.has_time():
            hour = max(min(hour, 23), 0)
            minute = max(min(minute, 59), 0)

            if second is not None:
                second = max(min(second, 60), 0) # support leap seconds

            self.data[3:6] = hour, minute, second

    def __str__(self):
        return Date.__str__(self) + self.time_string()

    def time_string(self):
        if self.has_time():
            data = self.as_tuple()
            time_str = " %02d:%02d" % data[3:5]
            if data[5] is not None:
                time_str += ":%02d" % data[5]
            if data[6] is not None:
                time_str += " %s" % data[6]
            return time_str
        else:
            return ""

    def as_datetime(self):
        return self

    def as_date(self):
        return Date(self.data[:3])

    def as_datetime_or_date(self):

        """
        Return a date for this datetime if fields are missing. Otherwise, return
        this datetime itself.
        """

        if not self.has_time():
            return self.as_date()
        else:
            return self

    def __cmp__(self, other):

        """
        The result of comparing this instance with 'other' is, if both instances
        are datetime instances, derived from a comparison of the datetimes
        converted to UTC. If one or both datetimes cannot be converted to UTC,
        the datetimes are compared using the basic temporal comparison which
        compares their raw time data.
        """

        this = self

        if this.has_time():
            if isinstance(other, DateTime):
                if other.has_time():
                    this_utc = this.to_utc()
                    other_utc = other.to_utc()
                    if this_utc is not None and other_utc is not None:
                        return cmp(this_utc.as_tuple(), other_utc.as_tuple())
                else:
                    other = other.padded()
        else:
            this = this.padded()

        return Date.__cmp__(this, other)

    def has_time(self):

        """
        Return whether this object has any time information. Objects without
        time information can refer to the very start of a day.
        """

        return self.data[3] is not None and self.data[4] is not None

    def time(self):
        return self.data[3:]

    def seconds(self):
        return self.data[5]

    def time_zone(self):
        return self.data[6]

    def set_time_zone(self, value):
        self.data[6] = value

    def padded(self, empty_value=0):

        """
        Return a datetime with missing fields defined as being the given
        'empty_value' or 0 if not specified.
        """

        data = []
        for x in self.data[:6]:
            if x is None:
                data.append(empty_value)
            else:
                data.append(x)

        data += self.data[6:]
        return DateTime(data)

    def to_utc(self):

        """
        Return this object converted to UTC, or None if such a conversion is not
        defined.
        """

        if not self.has_time():
            return None

        offset = self.utc_offset()
        if offset:
            hours, minutes = offset

            # Invert the offset to get the correction.

            hours, minutes = -hours, -minutes

            # Get the components.

            hour, minute, second, zone = self.time()
            date = self.as_date()

            # Add the minutes and hours.

            minute += minutes
            if minute < 0 or minute > 59:
                hour += minute / 60
                minute = minute % 60

            # NOTE: This makes various assumptions and probably would not work
            # NOTE: for general arithmetic.

            hour += hours
            if hour < 0:
                date = date.previous_day()
                hour += 24
            elif hour > 23:
                date = date.next_day()
                hour -= 24

            return date.as_datetime(hour, minute, second, "UTC")

        # Cannot convert.

        else:
            return None

    def utc_offset(self):

        "Return the UTC offset in hours and minutes."

        zone = self.time_zone()
        if not zone:
            return None

        # Support explicit UTC zones.

        if zone == "UTC":
            return 0, 0

        # Attempt to return a UTC offset where an explicit offset has been set.

        match = timezone_offset_regexp.match(zone)
        if match:
            if match.group("sign") == "-":
                sign = -1
            else:
                sign = 1

            hours = int(match.group("hours")) * sign
            minutes = int(match.group("minutes") or 0) * sign
            return hours, minutes

        # Attempt to handle Olson time zone identifiers.

        dt = self.as_olson_datetime()
        if dt:
            seconds = dt.utcoffset().seconds
            hours = seconds / 3600
            minutes = (seconds % 3600) / 60
            return hours, minutes

        # Otherwise return None.

        return None

    def olson_identifier(self):

        "Return the Olson identifier from any zone information."

        zone = self.time_zone()
        if not zone:
            return None

        # Attempt to match an identifier.

        match = timezone_olson_regexp.match(zone)
        if match:
            return match.group("olson")
        else:
            return None

    def _as_olson_datetime(self, hours=None):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier and the given 'hours' offset, raising one of
        the pytz exceptions in case of ambiguity.
        """

        olson = self.olson_identifier()
        if olson and pytz:
            tz = pytz.timezone(olson)
            data = self.padded().as_tuple()[:6]
            dt = datetime.datetime(*data)

            # With an hours offset, find a time probably in a previously
            # applicable time zone.

            if hours is not None:
                td = datetime.timedelta(0, hours * 3600)
                dt += td

            ldt = tz.localize(dt, None)

            # With an hours offset, adjust the time to define it within the
            # previously applicable time zone but at the presumably intended
            # position.

            if hours is not None:
                ldt -= td

            return ldt
        else:
            return None

    def as_olson_datetime(self):

        """
        Return a Python datetime object for this datetime interpreted using any
        Olson time zone identifier, choosing the time from the zone before the
        period of ambiguity.
        """

        try:
            return self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):

            # Try again, using an earlier local time and then stepping forward
            # in the chosen zone.
            # NOTE: Four hours earlier seems reasonable.

            return self._as_olson_datetime(-4)

    def ambiguous(self):

        "Return whether the time is local and ambiguous."

        try:
            self._as_olson_datetime()
        except (pytz.UnknownTimeZoneError, pytz.AmbiguousTimeError):
            return 1

        return 0

class Timespan(ActsAsTimespan):

    """
    A period of time which can be compared against others to check for overlaps.
    """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.start, self.end)

    def __hash__(self):
        return hash((self.start, self.end))

    def as_timespan(self):
        return self

    def as_limits(self):
        return self.start, self.end

    def is_before(self, a, b):

        """
        Return whether 'a' is before 'b'. Since the end datetime of one period
        may be the same as the start datetime of another period, and yet the
        first period is intended to be concluded by the end datetime and not
        overlap with the other period, a different test is employed for datetime
        comparisons.
        """

        # Datetimes without times can be equal to dates and be considered as
        # occurring before those dates.

        if isinstance(a, DateTime) and (isinstance(b, DateTime) or not a.has_time()):
            return a <= b
        else:
            return a < b

    def __contains__(self, other):

        """
        This instance is considered to contain 'other' if one is not before or
        after the other. If this instance overlaps or coincides with 'other',
        then 'other' is regarded as belonging to this instance's time period.
        """

        return self == other

    def __cmp__(self, other):

        """
        Return whether this timespan occupies the same period of time as the
        'other'. Timespans are considered less than others if their end points
        precede the other's start point, and are considered greater than others
        if their start points follow the other's end point.
        """

        if isinstance(other, ActsAsTimespan):
            other = other.as_timespan()

            if self.end is not None and other.start is not None and self.is_before(self.end, other.start):
                return -1
            elif self.start is not None and other.end is not None and self.is_before(other.end, self.start):
                return 1
            else:
                return 0

        else:
            if self.end is not None and self.is_before(self.end, other):
                return -1
            elif self.start is not None and self.is_before(other, self.start):
                return 1
            else:
                return 0

class TimespanCollection:

    """
    A collection of timespans providing a list-like interface supporting
    membership tests at a particular resolution.
    """

    def __init__(self, resolution, values=None):

        # Timespans need to be given converted start and end dates/times.

        if resolution == "date":
            self.convert_time = lambda x: x.as_date()
        elif resolution == "datetime":
            self.convert_time = lambda x: x.as_datetime_or_date()
        else:
            self.convert_time = lambda x: x

        self.values = values or []

    def convert(self, value):
        if isinstance(value, ActsAsTimespan):
            value = value.as_timespan()
            start, end = map(self.convert_time, value.as_limits())
            return Timespan(start, end)
        else:
            return self.convert_time(value)

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __getitem__(self, i):
        return self.values[i]

    def __setitem__(self, i, value):
        self.values[i] = value

    def __contains__(self, value):
        test_value = self.convert(value)
        return test_value in self.values

    def append(self, value):
        self.values.append(value)

    def insert(self, i, value):
        self.values.insert(i, value)

    def pop(self):
        return self.values.pop()

    def insert_in_order(self, value):
        bisect.insort_left(self, value)

def getCountry(s):

    "Find a country code in the given string 's'."

    match = country_code_regexp.search(s)

    if match:
        return match.group("code")
    else:
        return None

def getDate(s):

    "Parse the string 's', extracting and returning a date object."

    dt = getDateTime(s)
    if dt is not None:
        return dt.as_date()
    else:
        return None

def getDateTime(s):

    """
    Parse the string 's', extracting and returning a datetime object where time
    information has been given or a date object where time information is
    absent.
    """

    m = datetime_regexp.search(s)
    if m:
        groups = list(m.groups())

        # Convert date and time data to integer or None.

        return DateTime(map(int_or_none, groups[:6]) + [m.group("zone")]).as_datetime_or_date()
    else:
        return None

def getDateStrings(s):

    "Parse the string 's', extracting and returning all date strings."

    start = 0
    m = date_regexp.search(s, start)
    l = []
    while m:
        l.append("-".join(m.groups()))
        m = date_regexp.search(s, m.end())
    return l

def getMonth(s):

    "Parse the string 's', extracting and returning a month object."

    m = month_regexp.search(s)
    if m:
        return Month(map(int, m.groups()))
    else:
        return None

def getCurrentDate():

    "Return the current date as a (year, month, day) tuple."

    today = datetime.date.today()
    return Date((today.year, today.month, today.day))

def getCurrentMonth():

    "Return the current month as a (year, month) tuple."

    today = datetime.date.today()
    return Month((today.year, today.month))

def getCurrentYear():

    "Return the current year."

    today = datetime.date.today()
    return today.year

# User interface functions.

def getParameter(request, name, default=None):

    """
    Using the given 'request', return the value of the parameter with the given
    'name', returning the optional 'default' (or None) if no value was supplied
    in the 'request'.
    """

    return get_form(request).get(name, [default])[0]

def getQualifiedParameter(request, calendar_name, argname, default=None):

    """
    Using the given 'request', 'calendar_name' and 'argname', retrieve the
    value of the qualified parameter, returning the optional 'default' (or None)
    if no value was supplied in the 'request'.
    """

    argname = getQualifiedParameterName(calendar_name, argname)
    return getParameter(request, argname, default)

def getQualifiedParameterName(calendar_name, argname):

    """
    Return the qualified parameter name using the given 'calendar_name' and
    'argname'.
    """

    if calendar_name is None:
        return argname
    else:
        return "%s-%s" % (calendar_name, argname)

def getParameterDate(arg):

    "Interpret 'arg', recognising keywords and simple arithmetic operations."

    n = None

    if arg is None:
        return None

    elif arg.startswith("current"):
        date = getCurrentDate()
        if len(arg) > 8:
            n = int(arg[7:])

    elif arg.startswith("yearstart"):
        date = Date((getCurrentYear(), 1, 1))
        if len(arg) > 10:
            n = int(arg[9:])

    elif arg.startswith("yearend"):
        date = Date((getCurrentYear(), 12, 31))
        if len(arg) > 8:
            n = int(arg[7:])

    else:
        date = getDate(arg)

    if n is not None:
        date = date.day_update(n)

    return date

def getParameterMonth(arg):

    "Interpret 'arg', recognising keywords and simple arithmetic operations."

    n = None

    if arg is None:
        return None

    elif arg.startswith("current"):
        date = getCurrentMonth()
        if len(arg) > 8:
            n = int(arg[7:])

    elif arg.startswith("yearstart"):
        date = Month((getCurrentYear(), 1))
        if len(arg) > 10:
            n = int(arg[9:])

    elif arg.startswith("yearend"):
        date = Month((getCurrentYear(), 12))
        if len(arg) > 8:
            n = int(arg[7:])

    else:
        date = getMonth(arg)

    if n is not None:
        date = date.month_update(n)

    return date

def getFormDate(request, calendar_name, argname):

    """
    Return the date from the 'request' for the calendar with the given
    'calendar_name' using the parameter having the given 'argname'.
    """

    arg = getQualifiedParameter(request, calendar_name, argname)
    return getParameterDate(arg)

def getFormMonth(request, calendar_name, argname):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameter having the given 'argname'.
    """

    arg = getQualifiedParameter(request, calendar_name, argname)
    return getParameterMonth(arg)

def getFormDateTriple(request, yeararg, montharg, dayarg):

    """
    Return the date from the 'request' for the calendar with the given
    'calendar_name' using the parameters having the given 'yeararg', 'montharg'
    and 'dayarg' names.
    """

    year = getParameter(request, yeararg)
    month = getParameter(request, montharg)
    day = getParameter(request, dayarg)
    if year and month and day:
        return Date((int(year), int(month), int(day)))
    else:
        return None

def getFormMonthPair(request, yeararg, montharg):

    """
    Return the month from the 'request' for the calendar with the given
    'calendar_name' using the parameters having the given 'yeararg' and
    'montharg' names.
    """

    year = getParameter(request, yeararg)
    month = getParameter(request, montharg)
    if year and month:
        return Month((int(year), int(month)))
    else:
        return None

def getFullDateLabel(request, date):

    """
    Return the full month plus year label using the given 'request' and
    'year_month'.
    """

    if not date:
        return ""

    _ = request.getText
    year, month, day = date.as_tuple()[:3]
    start_weekday, number_of_days = date.month_properties()
    weekday = (start_weekday + day - 1) % 7
    day_label = _(getDayLabel(weekday))
    month_label = _(getMonthLabel(month))
    return "%s %s %s %s" % (day_label, day, month_label, year)

def getFullMonthLabel(request, year_month):

    """
    Return the full month plus year label using the given 'request' and
    'year_month'.
    """

    if not year_month:
        return ""

    _ = request.getText
    year, month = year_month.as_tuple()[:2]
    month_label = _(getMonthLabel(month))
    return "%s %s" % (month_label, year)

# Page-related functions.

def getPrettyPageName(page):

    "Return a nicely formatted title/name for the given 'page'."

    title = page.split_title(force=1)
    return getPrettyTitle(title)

def linkToPage(request, page, text, query_string=None):

    """
    Using 'request', return a link to 'page' with the given link 'text' and
    optional 'query_string'.
    """

    text = wikiutil.escape(text)
    return page.link_to_raw(request, text, query_string)

def getFullPageName(parent, title):

    """
    Return a full page name from the given 'parent' page (can be empty or None)
    and 'title' (a simple page name).
    """

    if parent:
        return "%s/%s" % (parent.rstrip("/"), title)
    else:
        return title

def fillEventPageFromTemplate(template_page, new_page, event_details, category_pagenames):

    """
    Using the given 'template_page', complete the 'new_page' by copying the
    template and adding the given 'event_details' (a dictionary of event
    fields), setting also the 'category_pagenames' to define category
    membership.
    """

    event_page = EventPage(template_page)
    new_event_page = EventPage(new_page)
    new_event_page.copyPage(event_page)

    if new_event_page.getFormat() == "wiki":
        new_event = Event(new_event_page, event_details)
        new_event_page.setEvents([new_event])
        new_event_page.setCategoryMembership(category_pagenames)
        new_event_page.saveChanges()

# vim: tabstop=4 expandtab shiftwidth=4
